var pull = require('pull-stream')
var multicb = require('multicb')
var crypto = require('crypto')
var cat = require('pull-cat')
var cache = require('pull-cache')
var ref = require('ssb-ref')
var createGitHash = require('pull-hash/ext/git')
var createSSBBlobHash = require('pull-hash/ext/ssb')
var gitPack = require('pull-git-pack')
var catPack = require('pull-git-pack-concat')
var GitRepo = require('pull-git-repo')
var Mentions = require('ssb-mentions')
var PullRequests = require('ssb-pull-requests')
var KVGraph = require('kvgraph')
var ssbGit = require('ssb-git')
var paramap = require('pull-paramap')

function readNext(fn) {
  var next, self = this
  return function (end, cb) {
    if (next) return next(end, cb)
    fn.call(self, function (err, _next) {
      if (err) return cb(err)
      next = _next
      next(null, cb)
    })
  }
}

module.exports = Repo

var issueUpdateRegex = /(?:[Cc]los(?:e[sd]?|ing)|[Ff]ix(?:e[sd]|ing)?) +(?:issue +)?(%[A-Za-z0-9\/+]{43}=\.sha256)/g
// msg id pattern from ssb-ref/index.js
// close/fix pattern based on https://github.com/gitlabhq/gitlabhq/blob/c5ed929d33981d4af25c701087a4eaca1348cc1b/doc/customization/issue_closing.md

var maxMsgSize = 8192

function getIssueMentions(object) {
  var ids = {}
  var mentions = []
  var text = object.title + '\n\n' + object.body
  issueUpdateRegex.lastIndex = 0
  for (var m; m = issueUpdateRegex.exec(text); ) {
    var id = m[1]
    if (id in ids) continue
    if (ref.isLink(id)) {
      mentions.push({
        link: id,
        open: false,
        object: object.id,
        label: object.title
      })
      ids[id] = true
    }
  }
  return mentions
}

function getBranches(sbot, root, cb) {
  var heads = {}, tails = {}
  heads[root] = true
  pull(
    sbot.links({dest: root, rel: 'branch', values: true}),
    pull.drain(function (link) {
      var key = link.key
      var c = link.value.content
      if (!c || !c.branch) return
      var branches = Array.isArray(c.branch) ? c.branch : [c.branch]
      for (var i = 0; i < branches.length; i++) {
        var branch = branches[i]
        if (branch in heads) {
          delete heads[branch]
        } else {
          tails[branch] = true
        }
      }
      if (key in tails) {
        delete tails[key]
      } else {
        heads[key] = true
      }
    }, function (err) {
      var branches = Object.keys(heads)
      cb(err, branches.length === 1 ? branches[0] : branches)
    })
  )
}

var emptyPackId = '&47hwmsDkBO4rXpJgiKY4dfJDoGB7oL/7wiimQsZL5wI=.sha256'
var emptyIdxId = '&JuEIZDf1XX38OXLTVlS8HCSXCD073j2AQP7ejQbgepc=.sha256'

var ignoreBlobs = (process.env.GITSSB_IGNORE_BLOBS || '').split(' ')

var w = 0
function getBlobWithTimeout(timeout, key, cb) {
  if (process.env.DEBUG_BLOBS)
    console.error(++w + '+', 'blob req', key)
  if (this.output)
    this.output.write('Fetching blob: ' + key + '\r')
  if (!key) return cb(new Error('Missing blob key'))

  if (~ignoreBlobs.indexOf(key)) timeout = 1

  var timer = setTimeout(function () {
    timer = null
    cb(new Error('Timed out waiting for blob \'' + key + '\''))
  }, timeout)
  this.sbot.blobs.want(key, function (err, got) {
    if (process.env.DEBUG_BLOBS)
      console.error(--w + '-', 'blob got', key)
    if (!timer) return console.error('Got blob after timeout', key, err || got)
    clearTimeout(timer)
    if (err) cb(err)
    else if (!got) cb(new Error('Unable to get blob ' + key))
    else cb(null, this.sbot.blobs.get(key))
  }.bind(this))
}

function hashGitObject(obj, cb) {
  return pull(obj.read, createGitHash(obj, cb), pull.drain())
}

function Repo(sbot, id, value, upstream, options) {
  this.sbot = sbot
  this.id = id
  this.feed = value.author
  this.upstream = upstream
  this.output = options.output
  this._blobs = {/* sha1: [Buffer] */}
  this._msgsByObject = {/* sha1: Msg */}
  this._packs = [/* {packId: blobId, idxId: blobId} */]
  this._mysteryPacks = []

  this.updates = new KVGraph('key')
  this.updates.add({key: id, value: value})

  // queued operations while syncing of old messages
  this._syncCbs = []

  this.pullReqs = PullRequests.init(this.sbot)

  // pull-git-repo uses these
  this.caches = options.caches
}

Repo.prototype.closed = false
Repo.prototype.synced = false

Repo.prototype.close = function (cb) {
  if (this.closed) return
  this.closed = true
  if (!cb) cb = function () {}
  var done = multicb({ pluck: 1 })
  if (this._readNew)
    this._readNew(true, done())
  if (this._readOld)
    this._readOld(true, done())
  this.pullReqs.deinit(cb)
  done(cb)
}

Repo.prototype.getHead = function (cb) {
  this.getState(function (err, state) {
    cb(null, state.head)
  })
}

function insert(array, item, atBeginning) {
  if (atBeginning) array.unshift(item)
  else array.push(item)
}

Repo.prototype._processMsg = function (msg, isNew) {
  var c = msg.value.content
  if (c.type !== 'git-update' || c.repo != this.id) return

  this.updates.add(msg)

  if (c.packs && c.indexes) for (var i = 0; i < c.packs.length; i++) {
    var packId = c.packs[i] && c.packs[i].link
    var idxId = c.indexes[i] && c.indexes[i].link
    if (packId) {
      var pack = {
        packId: packId,
        idxId: idxId,
        msgId: msg.key
      }
      insert(this._packs, pack, isNew)
      if (!c.object_ids) {
        // this pack might contain objects that the message doesn't mention
        insert(this._mysteryPacks, pack, isNew)
      }
    }
  }

  if (c.commits) for (var j = 0; j < c.commits.length; j++) {
    var commit = c.commits[j]
    if (commit && commit.sha1) {
      this._msgsByObject[commit.sha1] = msg
    }
  }

  if (c.object_ids) for (var k = 0; k < c.object_ids.length; k++) {
    this._msgsByObject[c.object_ids[k]] = msg
  }

  if (c.refs) for (var name in c.refs) {
    var hash = c.refs[name]
    if (!(hash in this._msgsByObject)) {
      this._msgsByObject[hash] = msg
    }
  }
}

Repo.prototype._processOldMsg = function (msg) {
  this._processMsg(msg, false)
}

Repo.prototype._processNewMsg = function (msg) {
  // expire cached state object
  delete this._state

  this._processMsg(msg, true)
}

Repo.prototype._getBlob = function (key, cb) {
  getBlobWithTimeout.call(this, 30e3, key, cb)
}

Repo.prototype._getBlobCachedOnce = function (key, cb) {
  // use cached blob if it exists (in case of newly pushed data)
  // but don't keep it cached because that wastes memory
  if (key in this._blobs) {
    var bufs = this._blobs[key]
    delete this._blobs[key]
    cb(null, pull.values(bufs))
  } else {
    this._getBlob(key, cb)
  }
}

// get refs source({name, hash})
Repo.prototype.refs = function (options) {
  if (!options) options = {}
  var inherited = options.inherited !== false

  return readNext.call(this, function (cb) {
    this.getState(function (err, state) {
      if (err) cb(err)
      else cb(null, pull(
        cat([
          pull.values(state.refs),
          inherited && this.upstream && this.upstream.refs()
        ]),
        pull.unique('name')
      ))
    })
  })
}

Repo.prototype.symrefs = function () {
  return readNext.call(this, function (cb) {
    this.getHead(function (err, head) {
      if (err) return cb(err)
      cb(null, pull.once({name: 'HEAD', ref: head}))
    })
  })
}

Repo.prototype._sync = function (live) {
  if (live)
    now = [Date.now()]

  if (this.synced) {
    // already synced
    return
  }

  pull(
    this._readOld = this.sbot.links({
      type: 'git-update',
      dest: this.id,
      // source: this.feed,
      rel: 'repo',
      values: true,
      reverse: true
    }),
    pull.drain(this._processOldMsg.bind(this), function (err) {
      // complete waiting requests for lookups
      this.synced = true
      while (this._syncCbs.length)
        this._syncCbs.pop().call(this)
      if (live)
        this._syncNew()
    }.bind(this))
  )
}

Repo.prototype.awaitSync = function (cb) {
  if (this.synced) cb.call(this)
  else this._syncCbs.push(cb)
}

Repo.prototype.getState = function (cb) {
  if (this._state) return cb.call(this, null, this._state)
  this.awaitSync(function () {
    cb.call(this, null, this._state = this.updates.reduceRight(ssbGit.merge))
  })
}

Repo.prototype._syncNew = function () {
  if (this.closed) return
  var id = this.id
  pull(
    this._readNew = this.sbot.createLogStream({
      values: true,
      live: true,
      old: false
    }),
    pull.filter(function (msg) {
      var c = msg.value.content
      return c && c.type == 'git-update' && c.repo == id
    }),
    pull.drain(this._processNewMsg.bind(this), function (err) {
      if (err) throw err
      delete this._readNew
    })
  )
}

Repo.prototype.hasObject = function (hash, cb) {
  this.awaitSync(function () {
    cb(this.hasObjectQuick(hash))
  })
}

Repo.prototype.hasObjectQuick = function (hash) {
  // false negative okay
  return !!this._msgsByObject[hash]
}

Repo.prototype.getObject = function (hash, cb) {
  cb()
}

Repo.prototype.getPackfile = function (id, cb) {
  this._getBlobCachedOnce(id, cb)
}

Repo.prototype.getPackIndex = function (id, cb) {
  this._getBlobCachedOnce(id, cb)
}

Repo.prototype.packs = function (hash) {
  if (!this.synced) return readNext.call(this, function (cb) {
    this.awaitSync(function () {
      cb(null, this.packs(hash))
    })
  })

  // Optimization: if we know what message published this hash,
  // return the pack for that first.
  if (hash) {
    var msg = this._msgsByObject[hash]
    if (msg) {
      var c = msg.value.content
      if (c) {
        var packId = c.packs && c.packs[0] && c.packs[0].link
        var idxId = c.indexes && c.indexes[0] && c.indexes[0].link
        if (packId && idxId) {
          var pack = {
            packId: packId,
            idxId: idxId,
            msgId: msg.key
          }
          return cat([
            pull.once(pack),
            pull.values(this._mysteryPacks)
          ])
        }
      }
    } else {
      // If a message has a object_ids array but it does not mention the
      // requested object, the message's pack probably doesn't contain our
      // object. So prioritize packs from messages that don't contain
      // object_ids.
      return pull.values(this._mysteryPacks)
    }
  }

  return pull.values(this._packs)
}

function Repo_readRefUpdates(updates, cb) {
  var self = this
  var refs
  var ended
  updates(null, function next(end, update) {
    if (ended = end) return cb(end === true ? null : end, refs)
    ;(refs || (refs = {}))[update.name] = update.new
    if (!ended)
      updates(null, next)
  })
}

Repo.prototype.uploadPack = function (readRefUpdates, readPacks, cb) {
  /*
  if (this.sbot.id && this.feed && this.sbot.id != this.feed)
    return cb(new Error('You do not have permission to push to this repo'))
    */
  var self = GitRepo(this)
  var done = multicb({ pluck: 1, spread: true })
  var packsCb = done()
  var indexesCb = done()
  var mentionsCb = done()
  var commitsCb = done()
  var refsCb = done()
  var repoBranchesCb = done()
  var commitsInfo = []
  var numCommitsPushed
  var numObjects
  var objectIds = []
  var linkedIds = {}
  var tempPackInfo

  function getObjectIds(commits) {
    var commitIds = {}
    commits.forEach(function (commit) {
      commitIds[commit.sha1] = true
    })
    return objectIds.filter(function (id) {
      return !commitIds[id]
    })
  }

  function objectIdsToMessages(objIds) {
    var easyMsgs = [/* msg */]
    var needsFetching = [/* obj id */]
    for (var id in objIds) {
      if (objectIds.indexOf(id) > -1) continue // skip object being pushed now
      var msg = self._msgsByObject[id]
      if (msg) easyMsgs.push(msg)
      else needsFetching.push(id)
    }
    return pull(
      cat([
        pull.values(easyMsgs),
        pull(
          pull.values(needsFetching),
          paramap(function (id, cb) {
            self.getPackedObjectOffsetCached(id, cb)
          }, 4),
          pull.map(function (pack) {
            var msgId = pack && pack.msgId
            if (!msgId) return console.error('Missing offset message id', pack)
            var msg = self.updates.objects[msgId]
            if (!msg) return console.error('missing message', msgId)
            return msg
          }),
          pull.filter()
        )
      ]),
      pull.unique('key')
    )
  }

  function filterHeads(getLinks, cb) {
    var heads = {}
    var tails = {}
    var objects = self.updates.objects
    var stack = []

    function addTail(id) {
      if (tails[id]) return
      tails[id] = true
      if (heads[id]) return delete heads[id]
      var msg = objects[id]
      if (!msg) {
        // a bug led to some messages that were not git-updates to be included
        // in repoBranches. TODO: add schema for ignoring these
        // console.error('missing object', id)
        return
      }
      addTails(msg)
    }

    function addTails(msg) {
      var branches = msg && getLinks(msg)
      //console.error("branches", msg.key, branches)
      if (branches && Array.isArray(branches)) {
        stack.push.apply(stack, branches)
      }
    }

    return pull.drain(function (msg) {
      if (tails[msg.key]) return
      heads[msg.key] = true
      addTails(msg)
      while (stack.length) {
        addTail(stack.pop())
      }
    }, function (err) {
      if (err) return cb(err)
      //console.log('link heads', Object.keys(heads))
      cb(null, Object.keys(heads))
    })
  }

  function getRepoBranches(ids, cb) {
    //console.error('links', Object.keys(ids))
    pull(
      objectIdsToMessages(ids),
      filterHeads(function (msg) {
        return msg.value.content.repoBranch
      }, cb)
    )
  }

  pull(
    readPacks,
    pull.asyncMap(function (pack, cb) {
      var done = multicb({ pluck: 1, spread: true })
      var packCached = cache(pack.pack)

      // add the pack and the idx to ssb
      pull(packCached(), self.addSSBBlob(done()))
      pull(pack.idx, self.addSSBBlob(done()))

      done(function (err, packId, idxId) {
        if (err) return cb(err)
        var done = multicb({ pluck: 1, spread: true })
        if (packId == emptyPackId && idxId == emptyIdxId) return cb()
        // add the pack and idx before publishing the message,
        // so that decoding deltas in the pack works
        tempPackInfo = self._mysteryPacks.unshift({
          packId: packId,
          idxId: idxId
        })

        // process packed objects
        var mentions = []
        var bar
        var objI = 0
        pull(
          packCached(),
          gitPack.decode({onHeader: function (_numObjects) {
            numObjects = _numObjects
          }}, self, done()),
          pull.asyncMap(function (obj, cb) {
            if (self.output) {
              self.output.write('Decoding objects: '
                + ((++objI / numObjects)*100).toFixed(0) + '%'
                + ' (' + objI  + '/' + numObjects + ')\r')
            }
            if (obj.type == 'commit')
              GitRepo.getCommitParsed(obj, cb)
            else if (obj.type == 'tag')
              GitRepo.getTagParsed(obj, cb)
            else if (obj.type == 'tree')
              GitRepo.getTreeParsed(obj, cb)
            else
              hashGitObject(obj, cb)
          }),
          pull.filter(),
          pull.drain(function (obj) {
            // collect object ids
            if (typeof obj === 'string') {
              objectIds.push(obj)
              return
            } else if (obj.id) {
              objectIds.push(obj.id)
            }

            // collect info about commits, for easier message rendering
            if (obj.committer) {
              commitsInfo.push({
                parents: obj.parents,
                title: obj.title,
                body: obj.body || undefined,
                commitDate: obj.committer.date,
                sha1: obj.id
              })
            }

            // record dependency info of git objects
            if (obj.entries) {
              // tree
              for (var i = 0; i < obj.entries.length; i++) {
                var item = obj.entries[i]
                if (item.mode !== 0160000) {
                  // skip link to git commit since it may not be in this repo
                  linkedIds[item.id] = true
                }
              }
            } else if (obj.committer) {
              // commit
              for (var i = 0; i < obj.parents.length; i++) {
                linkedIds[obj.parents[i]] = true
              }
              if (obj.tree) linkedIds[obj.tree] = true
            } else if (obj.tagger) {
              // tag
              if (obj.object) linkedIds[obj.object] = true
            }

            // detect mentions in commits and tags
            var text = obj.title + '\n\n' + obj.body
            Mentions(text).forEach(function (mention) {
              mention.object = obj.id
              mention.label = obj.title
              // TODO: add commit/tag title to mention?
              mentions.push(mention)
            })
          }, done())
        )

        done(function (err) {
          if (self.output) self.output.write('\n')
          if (err) return cb(err)
          cb(null, {
            pack: {link: packId},
            idx: {link: idxId},
            mentions: mentions,
          })
        })
      })
    }),
    pull.filter(Boolean),
    pull.collect(function (err, combined) {
      packsCb(err, combined.map(function (obj) { return obj.pack }))
      indexesCb(err, combined.map(function (obj) { return obj.idx }))
      mentionsCb(err, [].concat.apply([],
        combined.map(function (obj) { return obj.mentions })))

      // now that the packs are added to the repo, read the ref updates
      // and new commits
      if (readRefUpdates)
        Repo_readRefUpdates.call(self, readRefUpdates, gotRefUpdates)
      else
        gotRefUpdates()
    })
  )

  function gotRefUpdates(err, refUpdatesObj) {
    if (err) return refsCb(err)

    // Reify unmerged conflict branches
    self.getState(function (err, state) {
      if (err) return refsCb(err)

      // conflict branches automatically go away when a later ssb message
      // resolves the branch conflict. add a conflict branch explicitly here if
      // it was going to go away, so that commits are not lost.
      state.refs.forEach(function (ref) {
        if (ref.realname && ref.realname in refUpdatesObj) {
          if (refUpdatesObj[ref.name] === null) {
            delete refUpdatesObj[ref.name]
          } else if (!(ref.name in refUpdatesObj)) {
            refUpdatesObj[ref.name] = ref.hash
          }
        }
      })

      refsCb(null, refUpdatesObj)

      // Get messages for packfiles containing the linked objects
      getRepoBranches(linkedIds, repoBranchesCb)
    })

    // Detect issue close commands in commit messages,
    // and pull requests merged into HEAD.
    // Calculate which commits are being merged into HEAD.
    self.getHead(function (err, ref) {
      var newHead = refUpdatesObj[ref]
      if (!newHead) {
        // HEAD not updated
        return commitsCb(null, [])
      }

      // find common ancestor of old HEAD and new HEAD
      GitRepo.getMergeBase(self, ref, self, newHead, function (err, base) {
        if (err && err.name === 'NotFoundError') return commitsCb(null, [])
        else if (err) return commitsCb(err)
        // read commits from merge base to new HEAD
        pull(
          self.readLog(newHead),
          pull.take(function (rev) { return rev != base }),
          pull.asyncMap(self.getCommitParsed.bind(self)),
          pull.collect(commitsCb)
        )
      })
    })
  }

  // update the base branch of an open PR: check if the PR is merged
  var pullsByHead = {/* commitIdMerged: prToClose */}
  var pullsByHeadCb = done()
  pull(
    self.pullReqs.list({
      open: true,
      repo: self.id
    }),
    pull.asyncMap(function (pr, cb) {
      if (pr.headRepo == self.id) gotRepo(null, self)
      else require('../').getRepo(self.sbot, pr.headRepo, gotRepo)
      function gotRepo(err, headRepo) {
        if (err) return cb(err)
        GitRepo(headRepo).resolveRef(pr.headBranch, function (err, rev) {
          if (err && err.name === 'NotFoundError') rev = null
          else if (err) return cb(err)
          cb(null, { id: pr.id, headRev: rev })
        })
      }
    }),
    pull.filter(),
    pull.drain(function (pr) {
      pullsByHead[pr.headRev] = pr.id
    }, function (err) {
      pullsByHeadCb(err, pullsByHead)
    })
  )

  // update the head branch of an open PR: reply to the PR
  var pullsByBranch = {/* branchUpdated: prToReplyTo */}
  var pullsByBranchCb = done()
  pull(
    self.pullReqs.list({
      open: true,
      headRepo: self.id
    }),
    pull.drain(function (pr) {
      pullsByBranch[pr.headBranch] = pr.id
    }, function (err) {
      pullsByBranchCb(err, pullsByBranch)
    })
  )

  done(function (err, packs, indexes, mentions, commits, refUpdatesObj,
      repoBranches, pullsByHead, pullsByBranch) {
    if (err) return cb(err)
    if (packs.length == 0 && indexes.length == 0) packs = indexes = undefined

    // get issues and pull requests closed
    var issues = []
    commits.forEach(function (commit) {
      issues.push.apply(issues, getIssueMentions(commit))
      if (commit.id in pullsByHead) {
        var pullId = pullsByHead[commit.id]
        // PR closed because its head is one of the pushed commits
        issues.push({
          link: pullId,
          open: false,
          merged: true,
          object: commit.id,
          label: commit.title
        })
      }
    })

    if (issues && mentions) {
      // deduplicate closed issues and issue mentions
      var closed = {}
      issues.forEach(function (issue) {
        if (issue.open === false)
          closed[[issue.link, issue.object]] = true
      })
      mentions = mentions.filter(function (mention) {
        return !closed[[mention.link, mention.object]]
      })
    }

    if (mentions.length == 0) mentions = undefined
    if (issues.length == 0) issues = undefined

    // use commit date for sorting, but don't include it in the message
    commitsInfo.sort(function (a, b) {
      return b.commitDate - a.commitDate
    })
    commitsInfo.forEach(function (commitInfo) {
      delete commitInfo.commitDate
    })

    // don't need to include repo branches in refs branches
    var refsBranches = []
    for (var key in self.updates.getHeads()) {
      if (repoBranches.indexOf(key) == -1) refsBranches.push(key)
    }

    var msg = {
      type: 'git-update',
      repo: self.id,
      refsBranch: refsBranches,
      repoBranch: repoBranches.length > 0 ? repoBranches : undefined,
      packs: packs,
      indexes: indexes,
      refs: refUpdatesObj,
      mentions: mentions,
      issues: issues,
      commits: commitsInfo,
      num_objects: numObjects,
      object_ids: getObjectIds(commitsInfo),
    }

    // make a message a reply to a PR message if we are updating a PR branch
    for (var ref in refUpdatesObj) {
      var branch = ref.replace(/^refs\/heads\//, '')
      var prId = pullsByBranch[branch]
      if (prId) {
        msg.root = prId
        getBranches(self.sbot, prId, gotBranches)
        return
      }
    }
    gotBranches()

    function gotBranches(err, branch) {
      if (err) return cb(err)
      if (branch && branch.length > 0) msg.branch = branch
      numCommitsPushed = commitsInfo.length

      // push blobs
      var blobLinks = [].concat(packs || [], indexes || [])
      if (!blobLinks.length || !self.sbot.blobs.push || process.env.DRY_RUN) {
        setImmediate(pushedBlobs)
      } else {
        console.error('Pushing blobs...')
        var done = multicb({ pluck: 1 })
        blobLinks.forEach(function (link) {
          self.sbot.blobs.push(link.link, done())
        })
        done(pushedBlobs)
      }
    }

    function pushedBlobs(err) {
      if (err) return cb(err)
      tryPublish(msg)
    }
  })

  function tryPublish(value) {
    if (process.env.DRY_RUN) {
      console.error(JSON.stringify(value, 0, 2))
      process.exit(1)
    }

    self.sbot.publish(value, function (err, msg) {
      if (err) {
        if (/must not be large/.test(err.message)) {
          // try to make the message fit under the size limit
          if (value.commits.length > 0) {
            // truncate commits array
            value.commits_more = numCommitsPushed - (value.commits.length >>= 1)
            // update object_ids to include ids removed from the commits array
            value.object_ids = getObjectIds(value.commits)
            tryPublish(value)
          } else if (value.object_ids) {
            // give up on including object_ids
            delete value.object_ids
            tryPublish(value)
          } else {
            cb(new Error('Message is too big to send'))
          }
        } else {
          cb(err)
        }
      } else {
        console.error('Published ' + msg.key)

        self._processNewMsg(msg)

        // remove pack info from before we knew the id of the message
        var i = self._mysteryPacks.indexOf(tempPackInfo)
        if (i !== -1) self._mysteryPacks.splice(i, 1)

        // done
        cb()
      }
    })
  }
}

Repo.prototype.addSSBBlob = function (cb) {
  // work around sbot.blobs.add not calling back with blob id
  var done = multicb({ pluck: 1, spread: true })
  var hashCb = done()
  var bufs = []
  var self = this
  done(function (err, hash) {
    if (err) return cb(err)
    // add the blob locally so we don't have to request it again from sbot
    self._blobs[hash] = bufs
    cb(null, hash)
  })
  return pull(
    createSSBBlobHash(hashCb),
    pull.through(function (buf) {
      bufs.push(buf)
    }),
    this.sbot.blobs.add(done())
  )
}

function fetchBlobs(blobs, blobIds, output, cb) {
  var aborted
  var fetched = 0
  if (blobIds.length === 0) return cb()

  function printProgress() {
    output.write('Fetching blobs: ' + fetched + '/' + blobIds.length + '\r')
  }
  if (output) printProgress()

  blobIds.forEach(function (blobId) {
    if (aborted) return
    blobs.want(blobId, function (err, got) {
      if (aborted) return
      if (err) return cb(aborted = err)
      if (!got) return cb(aborted = new Error('Missing blob ' + blobId))
      fetched++
      if (output) printProgress()
      if (fetched === blobIds.length) cb()
    })
  })
}

if (!process.env.GITSSB_OLD_PACKS)
Repo.prototype.getPack = function (wants, haves, cb) {
  // ensure our id is present
  if (!this.sbot.id) return this.sbot.whoami(function (err, feed) {
    if (err) return cb(err)
    this.sbot.id = feed.id
    this.getPack(wants, haves, cb)
  }.bind(this))

  // construct a packfile containing all objects needed to update from haves to
  // wants. instead of walking the commit graph, walk the ssb msg graph.
  var packs = []
  var queue = []
  var getHeads = false

  // map want-commits to want-msgs
  var wantMsgs = {}
  for (var sha1 in wants) {
    for (var repo = this, msg = null; repo && !msg; repo = repo.upstream)
      msg = repo._msgsByObject[sha1]
    if (msg) {
      wantMsgs[msg.key] = true
      if (!msg.value.content.repoBranch) {
        // old repo without repoBranch: add heads which aren't yet linked to
        getHeads = true
      }
    }
  }
  if (getHeads) {
    for (var repo = this; repo; repo = repo.upstream) {
      for (var key in repo.updates.getHeads()) {
        wantMsgs[key] = true
      }
    }
  }

  // map have-commits to have-msgs
  var haveMsgs = {}
  for (var sha1 in haves) {
    for (var repo = this, msg = null; repo && !msg; repo = repo.upstream)
      msg = repo._msgsByObject[sha1]
    if (msg) {
      // don't let own messages interfere. we may have content that we pushed
      // which links to other content that we don't have,
      // if we pushed something without fetching first.
      if (msg.value.author !== this.sbot.id) {
        haveMsgs[msg.key] = true
      }
    }
  }

  // special case: if a want msg is also a have msg, ignore the have.
  // this may happen if the repo was updated from another remote
  // TODO: make this better
  for (var key in wantMsgs) {
    if (key in haveMsgs) {
      delete haveMsgs[key]
    }
  }

  // walk from want-msgs back to the have-msgs
  for (var key in wantMsgs) queue.push(key)
  var seen = {}
  var blobIds = []
  while (queue.length) {
    var key = queue.shift()
    if (!key || key in haveMsgs || key in seen) continue
    seen[key] = true
    for (repo = this, msg = null; repo && !msg; repo = repo.upstream)
      msg = repo.updates.objects[key]
    var c = msg && msg.value && msg.value.content
    if (c && c.type === 'git-update') {
      var packId = c.packs && c.packs[0] && c.packs[0].link
      var idxId = c.indexes && c.indexes[0] && c.indexes[0].link
      if (packId && idxId) {
        packs.push({
          packId: packId,
          idxId: idxId
        })
        blobIds.push(packId, idxId)
      }
      if (Array.isArray(c.repoBranch)) {
        queue.push.apply(queue, c.repoBranch)
      }
      if (!c.refsBranch) {
        // old schema followed properties other than repoBranch for git object
        // links
        queue.push(msg.value.previous)
        queue.push(c.branch)
      }
    }
  }

  // want() the packs and pack-indexes
  var blobs = this.sbot.blobs
  fetchBlobs(blobs, blobIds, this.output, function (err) {
    if (err) return cb(aborted = err)
    packs.forEach(function (pack) {
      pack.read = blobs.get(pack.packId)
      pack.readIdx = blobs.get(pack.idxId)
    })
    // concatenate and deduplicate the packs
    cb(null, catPack(packs))
  })
}

Repo.prototype.getRefLog = function (ref, options, cb) {
  if (/^[0-9a-f]{40}$/.test(ref)) return cb.call(this, null, [ref])

  this.awaitSync(function () {
    var headRef = 'refs/heads/' + ref
    var tagRef = 'refs/tags/' + ref
    cb.call(this, null, this.updates.reduceLeft(ssbGit.linearize)
      .map(function (msg) {
        var c = msg && msg.value && msg.value.content
        var refs = c && c.refs
        return refs && (refs[ref] || refs[headRef] || refs[tagRef])
      }).filter(Boolean))
  }.bind(this))
}

Repo.prototype.getLatestAvailableRev = function (ref, timeout, cb) {
  this.getRefLog(ref, null, function (err, revs) {
    if (err) return cb(err)
    pull(
      pull.values(revs),
      GitRepo(this).getFirstAvailableRev(10e3, function (err, rev) {
        if (err === true) return cb(null, null, revs.length)
        if (err) return cb(err)
        cb(null, rev, revs.indexOf(rev))
      })
    )
  })
}
